(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('@angular/core'), require('@angular/common'), require('@angular/forms'), require('@ngrx/store'), require('rxjs/operators'), require('@ngrx/effects'), require('rxjs/observable/of'), require('rxjs/Observable')) :
	typeof define === 'function' && define.amd ? define(['exports', '@angular/core', '@angular/common', '@angular/forms', '@ngrx/store', 'rxjs/operators', '@ngrx/effects', 'rxjs/observable/of', 'rxjs/Observable'], factory) :
	(factory((global['ngrx-store-forms'] = {}),global.ng.core,global.ng.common,global.ng.forms,global.ngrx.store,global.Rx.Observable.prototype,global.ngrx.effects,global.Rx.Observable,global.Rx));
}(this, (function (exports,core,common,forms,store,operators,effects,of,Observable) { 'use strict';

/**
 * @return {?}
 */
function noStoreError() {
    throw new Error("No store detected. Make sure you've included @ngrx/store in your application.");
}
/**
 * @return {?}
 */
function noFormGroupError() {
    throw new Error("rxfConnectForm needs to be placed on an element which also contains a formGroup directive.");
}
/**
 * @param {?} path
 * @return {?}
 */
function noStoreFormBinding(path) {
    throw new Error("No binding found for form state on path '" + path + "'.");
}
var STORE_FORMS_CONFIG = new core.InjectionToken('rxsf Config');
/**
 * @param {?} object
 * @param {?} path
 * @return {?}
 */
function deepGet(object, path) {
    return path
        .split('.')
        .reduce(function (partialState, pathSegment) { return partialState[pathSegment] || {}; }, object);
}
/**
 * @param {?} x
 * @param {?} y
 * @return {?}
 */
function deepEquals(x, y) {
    if (x === y) {
        return true; // if both x and y are null or undefined and exactly the same
    }
    else if (!(x instanceof Object) || !(y instanceof Object)) {
        return false; // if they are not strictly equal, they both need to be Objects
    }
    else if (x.constructor !== y.constructor) {
        // they must have the exact same prototype chain, the closest we can do is
        // test their constructor.
        return false;
    }
    else {
        for (var /** @type {?} */ p in x) {
            if (!x.hasOwnProperty(p)) {
                continue; // other properties were tested using x.constructor === y.constructor
            }
            if (!y.hasOwnProperty(p)) {
                return false; // allows to compare x[ p ] and y[ p ] when set to undefined
            }
            if (x[p] === y[p]) {
                continue; // if they have the same strict value or identity then they are equal
            }
            if (typeof (x[p]) !== 'object') {
                return false; // Numbers, Strings, Functions, Booleans must be strictly equal
            }
            if (!deepEquals(x[p], y[p])) {
                return false;
            }
        }
        for (var /** @type {?} */ p in y) {
            if (y.hasOwnProperty(p) && !x.hasOwnProperty(p)) {
                return false;
            }
        }
        return true;
    }
}
var UpdateStoreFormStateAction = /** @class */ (function () {
    /**
     * @param {?} path
     * @param {?} formState
     */
    function UpdateStoreFormStateAction(path, formState) {
        this.path = path;
        this.formState = formState;
        this.type = 'UpdateStoreFormStateAction';
    }
    return UpdateStoreFormStateAction;
}());
var UpdateStoreFormAction = /** @class */ (function () {
    /**
     * @param {?} path
     * @param {?} value
     */
    function UpdateStoreFormAction(path, value) {
        this.path = path;
        this.value = value;
        this.type = 'UpdateStoreFormAction';
    }
    return UpdateStoreFormAction;
}());
var StoreFormsNoopAction = /** @class */ (function () {
    function StoreFormsNoopAction() {
        this.type = 'StoreFormsNoopAction';
    }
    return StoreFormsNoopAction;
}());
/**
 * @param {?} state
 * @param {?} action
 * @return {?}
 */
function storeFormsReducer(state, action) {
    switch (action.type) {
        case 'UpdateStoreFormStateAction': {
            var /** @type {?} */ clone = Object.assign({}, state);
            var /** @type {?} */ formState = action.path
                .split('.')
                .reduce(function (partialState, pathSegment) {
                return partialState[pathSegment] = Object.assign({}, partialState[pathSegment]);
            }, clone);
            Object.assign(formState, Object.assign({}, action.formState));
            if (action.formState.value) {
                Object.assign(formState, {
                    value: Object.assign({}, formState.value, action.formState.value)
                });
            }
            return clone;
        }
    }
    return state;
}
/**
 * @param {?} reducer
 * @return {?}
 */
function storeFormsMetaReducer(reducer) {
    return function (state, action) {
        var /** @type {?} */ updatedState = storeFormsReducer(state, action);
        return reducer(updatedState, action);
    };
}
var BindingService = /** @class */ (function () {
    /**
     * @param {?} config
     * @param {?} store
     */
    function BindingService(config, store$$1) {
        this.config = config;
        this.store = store$$1;
        this.bindings = {};
        if (!this.store) {
            noStoreError();
        }
    }
    /**
     * @param {?} path
     * @param {?} formGroup
     * @return {?}
     */
    BindingService.prototype.bind = function (path, formGroup) {
        var _this = this;
        var /** @type {?} */ formGroupSubscription = formGroup.statusChanges
            .pipe(operators.startWith(true))
            .subscribe(function () {
            _this.store.dispatch(new UpdateStoreFormStateAction(path, {
                value: Object.assign({}, formGroup.getRawValue()),
                untouched: formGroup.untouched,
                touched: formGroup.touched,
                pristine: formGroup.pristine,
                dirty: formGroup.dirty,
                valid: formGroup.valid,
                invalid: formGroup.invalid,
                pending: formGroup.pending,
                errors: _this.getErrors(path, formGroup)
            }));
        });
        this.bindings[path] = {
            path: path,
            formGroup: formGroup,
            formGroupSubscription: formGroupSubscription
        };
        if (this.config.bindingStrategy === 'ObserveStore') {
            this.bindings[path].storeSubscription = this.store
                .pipe(operators.map(function (state) { return deepGet(state, path); }), operators.filter(function (formState) {
                // Very simple dirty checking by comparing current values in
                // state to values in form group using deep equal to prevent
                // infinite loop setValue -> store -> setValue -> store ...
                return formState &&
                    formState.value &&
                    Object.keys(formState.value).length !== 0 &&
                    !deepEquals(formState.value, formGroup.getRawValue());
            }))
                .subscribe(function (formState) {
                formGroup.setValue(formState.value);
            });
        }
    };
    /**
     * @param {?} path
     * @param {?} formGroup
     * @return {?}
     */
    BindingService.prototype.getErrors = function (path, formGroup) {
        if (!this.config.errorMessages) {
            return {};
        }
        var /** @type {?} */ errorMessages = deepGet(this.config.errorMessages, path);
        return Object.keys(formGroup.controls).reduce(function (groupErrors, controlName) {
            var /** @type {?} */ control = formGroup.controls[controlName];
            if (control.errors) {
                groupErrors[controlName] = Object.keys(control.errors).reduce(function (errors, validatorKey) {
                    if (control.errors &&
                        control.errors[validatorKey] &&
                        errorMessages[controlName] && errorMessages[controlName][validatorKey]) {
                        var /** @type {?} */ errorMessage = errorMessages[controlName][validatorKey];
                        if (errors.indexOf(errorMessage) === -1) {
                            errors.push(errorMessage);
                        }
                    }
                    return errors;
                }, /** @type {?} */ ([]));
            }
            return groupErrors;
        }, {});
    };
    /**
     * @param {?} path
     * @return {?}
     */
    BindingService.prototype.unbind = function (path) {
        this.bindings[path].formGroupSubscription.unsubscribe();
        if (this.bindings[path].storeSubscription) {
            this.bindings[path].storeSubscription.unsubscribe();
        }
        delete this.bindings[path];
    };
    /**
     * @param {?} path
     * @param {?} value
     * @return {?}
     */
    BindingService.prototype.updateFormGroup = function (path, value) {
        if (!this.bindings[path]) {
            noStoreFormBinding(path);
        }
        var /** @type {?} */ binding = this.bindings[path];
        binding.formGroup.patchValue(value);
    };
    return BindingService;
}());
BindingService.decorators = [
    { type: core.Injectable },
];
/**
 * @nocollapse
 */
BindingService.ctorParameters = function () { return [
    { type: undefined, decorators: [{ type: core.Inject, args: [STORE_FORMS_CONFIG,] },] },
    { type: store.Store, decorators: [{ type: core.Optional },] },
]; };
var BindingDirective = /** @class */ (function () {
    /**
     * @param {?} formGroupDirective
     * @param {?} store
     * @param {?} bindingService
     */
    function BindingDirective(formGroupDirective, store$$1, bindingService) {
        this.formGroupDirective = formGroupDirective;
        this.store = store$$1;
        this.bindingService = bindingService;
        if (!this.formGroupDirective) {
            noFormGroupError();
        }
        if (!this.store) {
            noStoreError();
        }
    }
    /**
     * @return {?}
     */
    BindingDirective.prototype.ngOnInit = function () {
        this.bindingService.bind(this.rxsfBinding, this.formGroupDirective.form);
    };
    /**
     * @return {?}
     */
    BindingDirective.prototype.ngOnDestroy = function () {
        this.bindingService.unbind(this.rxsfBinding);
    };
    return BindingDirective;
}());
BindingDirective.decorators = [
    { type: core.Directive, args: [{
                selector: '[rxsfBinding]'
            },] },
];
/**
 * @nocollapse
 */
BindingDirective.ctorParameters = function () { return [
    { type: forms.FormGroupDirective, decorators: [{ type: core.Self }, { type: core.Optional },] },
    { type: store.Store, decorators: [{ type: core.Optional },] },
    { type: BindingService, },
]; };
BindingDirective.propDecorators = {
    'rxsfBinding': [{ type: core.Input },],
};
var defaultStoreFormsConfig = {
    bindingStrategy: 'ObserveStore'
};
var StoreFormsModule = /** @class */ (function () {
    function StoreFormsModule() {
    }
    /**
     * @param {?=} config
     * @return {?}
     */
    StoreFormsModule.configure = function (config) {
        return {
            ngModule: StoreFormsModule,
            providers: [{
                    provide: STORE_FORMS_CONFIG,
                    useValue: Object.assign({}, defaultStoreFormsConfig, config)
                }, {
                    provide: BindingService,
                    useClass: BindingService,
                    deps: [STORE_FORMS_CONFIG, store.Store]
                }]
        };
    };
    return StoreFormsModule;
}());
StoreFormsModule.decorators = [
    { type: core.NgModule, args: [{
                imports: [
                    common.CommonModule,
                    forms.FormsModule,
                    forms.ReactiveFormsModule
                ],
                declarations: [
                    BindingDirective
                ],
                exports: [
                    BindingDirective
                ],
                providers: [{
                        provide: STORE_FORMS_CONFIG,
                        useValue: defaultStoreFormsConfig
                    }, {
                        provide: BindingService,
                        useClass: BindingService,
                        deps: [STORE_FORMS_CONFIG, store.Store]
                    }]
            },] },
];
/**
 * @nocollapse
 */
StoreFormsModule.ctorParameters = function () { return []; };
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function")
        r = Reflect.decorate(decorators, target, key, desc);
    else
        for (var i = decorators.length - 1; i >= 0; i--)
            if (d = decorators[i])
                r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function")
        return Reflect.metadata(k, v);
};
var StoreFormsEffects = /** @class */ (function () {
    /**
     * @param {?} actions
     * @param {?} bindingService
     * @param {?} config
     * @param {?} store
     */
    function StoreFormsEffects(actions, bindingService, config, store$$1) {
        var _this = this;
        this.actions = actions;
        this.bindingService = bindingService;
        this.config = config;
        this.store = store$$1;
        this.reflectToFormStateEffect = this.actions
            .pipe(operators.filter(function (action) { return action instanceof UpdateStoreFormAction; }), operators.switchMap(function (action) {
            if (_this.config.bindingStrategy === 'ObserveStore') {
                return /** @type {?} */ (_this.store
                    .pipe(operators.take(1), operators.map(function (state) {
                    var /** @type {?} */ formState = deepGet(state, action.path);
                    return new UpdateStoreFormStateAction(action.path, Object.assign({}, formState, { value: Object.assign({}, formState.value, action.value) }));
                })));
            }
            else {
                _this.bindingService.updateFormGroup(action.path, action.value);
                return of.of(new StoreFormsNoopAction());
            }
        }));
    }
    return StoreFormsEffects;
}());
StoreFormsEffects.decorators = [
    { type: core.Injectable },
];
/**
 * @nocollapse
 */
StoreFormsEffects.ctorParameters = function () { return [
    { type: effects.Actions, },
    { type: BindingService, },
    { type: undefined, decorators: [{ type: core.Inject, args: [STORE_FORMS_CONFIG,] },] },
    { type: store.Store, },
]; };
__decorate([
    effects.Effect(),
    __metadata("design:type", Observable.Observable)
], StoreFormsEffects.prototype, "reflectToFormStateEffect", void 0);

exports.StoreFormsModule = StoreFormsModule;
exports.BindingDirective = BindingDirective;
exports.BindingService = BindingService;
exports.StoreFormsEffects = StoreFormsEffects;
exports.storeFormsMetaReducer = storeFormsMetaReducer;
exports.UpdateStoreFormAction = UpdateStoreFormAction;
exports.ɵa = defaultStoreFormsConfig;
exports.ɵb = STORE_FORMS_CONFIG;

Object.defineProperty(exports, '__esModule', { value: true });

})));
//# sourceMappingURL=ngrx-store-forms.umd.js.map
